# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides a wrapper to aid in plotting."""
from beartype.typing import Any, Dict, List, Optional, Union
import numpy as np
import pyvista as pv

from ansys.geometry.core.designer.body import Body, MasterBody
from ansys.geometry.core.designer.face import Face
from ansys.geometry.core.logger import LOG as logger
from ansys.geometry.core.plotting.plotter import (
    DEFAULT_COLOR,
    EDGE_COLOR,
    PICKED_COLOR,
    PICKED_EDGE_COLOR,
    Plotter,
)
from ansys.geometry.core.plotting.plotting_types import EdgePlot, GeomObjectPlot
from ansys.geometry.core.plotting.trame_gui import _HAS_TRAME, TrameVisualizer
from ansys.geometry.core.plotting.widgets import (
    CameraPanDirection,
    DisplacementArrow,
    MeasureWidget,
    PlotterWidget,
    Ruler,
    ShowDesignPoints,
    ViewButton,
    ViewDirection,
)
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.sketch.sketch import Sketch


class PlotterHelper:
    """
    Provides for simplifying the selection of trame in ``plot()`` functions.

    Parameters
    ----------
    use_trame : bool, default: None
        Whether to enable the use of `trame <https://kitware.github.io/trame/index.html>`_.
        The default is ``None``, in which case the ``USE_TRAME`` global setting
        is used.
    allow_picking: bool, default: False
        Enables/disables the picking capabilities in the PyVista plotter.
    """

    def __init__(
        self, use_trame: Optional[bool] = None, allow_picking: Optional[bool] = False
    ) -> None:
        """Initialize ``use_trame`` and save current ``pv.OFF_SCREEN`` value."""
        # Check if the use of trame was requested
        if use_trame is None:
            import ansys.geometry.core as pyansys_geometry

            use_trame = pyansys_geometry.USE_TRAME

        self._use_trame = use_trame
        self._allow_picking = allow_picking
        self._pv_off_screen_original = bool(pv.OFF_SCREEN)
        self._geom_object_actors_map = {}
        self._pl = None
        self._picked_list = set()
        self._picker_added_actors_map = {}
        self._edge_actors_map = {}
        self._widgets = []

        if self._use_trame and _HAS_TRAME:
            # avoids GUI window popping up
            pv.OFF_SCREEN = True
            self._pl = Plotter(enable_widgets=False)
        elif self._use_trame and not _HAS_TRAME:
            warn_msg = (
                "'use_trame' is active but trame dependencies are not installed."
                "Consider installing 'pyvista[trame]' to use this functionality."
            )
            logger.warning(warn_msg)
            self._pl = Plotter()
        else:
            self._pl = Plotter()

        self._enable_widgets = self._pl._enable_widgets

    def enable_widgets(self):
        """Enable the widgets for the plotter."""
        # Create Plotter widgets
        if self._enable_widgets:
            self._widgets: List[PlotterWidget] = []
            self._widgets.append(Ruler(self._pl._scene))
            [
                self._widgets.append(DisplacementArrow(self._pl._scene, direction=dir))
                for dir in CameraPanDirection
            ]
            [
                self._widgets.append(ViewButton(self._pl._scene, direction=dir))
                for dir in ViewDirection
            ]
            self._widgets.append(MeasureWidget(self))
            self._widgets.append(ShowDesignPoints((self)))

    def select_object(self, geom_object: Union[GeomObjectPlot, EdgePlot], pt: np.ndarray) -> None:
        """
        Select an object in the plotter.

        Highlights the object edges and adds a label with the object name and adds
        it to the PyAnsys Geometry object selection.

        Parameters
        ----------
        geom_object : Union[GeomObjectPlot, EdgePlot]
            Geometry object to select.
        pt : ~numpy.ndarray
            Set of points to determine the label position.
        """
        added_actors = []

        # Add edges if selecting an object
        if isinstance(geom_object, GeomObjectPlot):
            geom_object.actor.prop.color = PICKED_COLOR
            children_list = geom_object.edges
            for edge in children_list:
                edge.actor.SetVisibility(True)
                edge.actor.prop.color = EDGE_COLOR
        elif isinstance(geom_object, EdgePlot):
            geom_object.actor.prop.color = PICKED_EDGE_COLOR

        text = geom_object.name

        label_actor = self._pl.scene.add_point_labels(
            [pt],
            [text],
            always_visible=True,
            point_size=0,
            render_points_as_spheres=False,
            show_points=False,
        )
        if geom_object.name not in self._picked_list:
            self._picked_list.add(geom_object.name)
        added_actors.append(label_actor)
        self._picker_added_actors_map[geom_object.actor.name] = added_actors

    def unselect_object(self, geom_object: Union[GeomObjectPlot, EdgePlot]) -> None:
        """
        Unselect an object in the plotter.

        Removes edge highlighting and label from a plotter actor and removes it
        from the PyAnsys Geometry object selection.

        Parameters
        ----------
        geom_object : Union[GeomObjectPlot, EdgePlot]
            Object to unselect.
        """
        # remove actor from picked list and from scene
        object_name = geom_object.name
        if object_name in self._picked_list:
            self._picked_list.remove(object_name)

        if isinstance(geom_object, GeomObjectPlot):
            geom_object.actor.prop.color = DEFAULT_COLOR
        elif isinstance(geom_object, EdgePlot):
            geom_object.actor.prop.color = EDGE_COLOR

        if geom_object.actor.name in self._picker_added_actors_map:
            self._pl.scene.remove_actor(self._picker_added_actors_map[geom_object.actor.name])

            # remove actor and its children(edges) from the scene
            if isinstance(geom_object, GeomObjectPlot):
                for edge in geom_object.edges:
                    # hide edges in the scene
                    edge.actor.SetVisibility(False)
                    # recursion
                    self.unselect_object(edge)
            self._picker_added_actors_map.pop(geom_object.actor.name)

    def picker_callback(self, actor: "pv.Actor") -> None:
        """
        Define callback for the element picker.

        Parameters
        ----------
        actor : ~pyvista.Actor
            Actor that we are picking.
        """
        pt = self._pl.scene.picked_point

        # if object is a body/component
        if actor in self._geom_object_actors_map:
            body_plot = self._geom_object_actors_map[actor]
            if body_plot.object.name not in self._picked_list:
                self.select_object(body_plot, pt)
            else:
                self.unselect_object(body_plot)

        # if object is an edge
        elif actor in self._edge_actors_map and actor.GetVisibility():
            edge = self._edge_actors_map[actor]
            if edge.name not in self._picked_list:
                self.select_object(edge, pt)
            else:
                self.unselect_object(edge)
                actor.prop.color = EDGE_COLOR

    def compute_edge_object_map(self) -> Dict[pv.Actor, EdgePlot]:
        """
        Compute the mapping between plotter actors and ``EdgePlot`` objects.

        Returns
        -------
        Dict[~pyvista.Actor, EdgePlot]
            Mapping between plotter actors and EdgePlot objects.
        """
        for object in self._geom_object_actors_map.values():
            # get edges only from bodies
            geom_obj = object.object
            if (
                isinstance(geom_obj, Body)
                or isinstance(geom_obj, MasterBody)
                or isinstance(geom_obj, Face)
                or isinstance(geom_obj, SketchFace)
                or isinstance(geom_obj, Sketch)
            ):
                for edge in object.edges:
                    self._edge_actors_map[edge.actor] = edge

    def enable_picking(self):
        """Enable picking capabilities in the plotter."""
        self._pl.scene.enable_mesh_picking(
            callback=self.picker_callback, use_actor=True, show=False, show_message=False
        )

    def disable_picking(self):
        """Disable picking capabilities in the plotter."""
        self._pl.scene.disable_picking()

    def add(
        self,
        object: Any,
        merge_bodies: bool = False,
        merge_component: bool = False,
        filter: str = None,
        **plotting_options,
    ):
        """
        Add a ``pyansys-geometry`` or ``PyVista`` object to the plotter.

        Parameters
        ----------
        object : Any
            Object or list of objects you want to show.
        merge_bodies : bool, default: False
            Whether to merge each body into a single dataset. When ``True``,
            all the faces of each individual body are effectively combined
            into a single dataset without separating faces.
        merge_component : bool, default: False
            Whether to merge this component into a single dataset. When ``True``,
            all the individual bodies are effectively combined into a single
            dataset without any hierarchy.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.
        """
        if isinstance(object, List) and not isinstance(object[0], pv.PolyData):
            logger.debug("Plotting objects in list...")
            self._pl.add_list(object, merge_bodies, merge_component, filter, **plotting_options)
        else:
            self._pl.add(object, merge_bodies, merge_component, filter, **plotting_options)

    def plot(
        self,
        object: Any = None,
        screenshot: Optional[str] = None,
        merge_bodies: bool = False,
        merge_component: bool = False,
        view_2d: Dict = None,
        filter: str = None,
        **plotting_options,
    ) -> List[Any]:
        """
        Plot and show any PyAnsys Geometry object.

        These types of objects are supported: ``Body``, ``Component``, ``List[pv.PolyData]``,
        ``pv.MultiBlock``, and ``Sketch``.

        Parameters
        ----------
        object : Any, default: None
            Any object or list of objects that you want to plot.
        screenshot : str, default: None
            Path for saving a screenshot of the image that is being represented.
        merge_bodies : bool, default: False
            Whether to merge each body into a single dataset. When ``True``,
            all the faces of each individual body are effectively combined
            into a single dataset without separating faces.
        merge_component : bool, default: False
            Whether to merge this component into a single dataset. When ``True``,
            all the individual bodies are effectively combined into a single
            dataset without any hierarchy.
        view_2d : Dict, default: None
            Dictionary with the plane and the viewup vectors of the 2D plane.
        filter : str, default: None
            Regular expression with the desired name or names you want to include in the plotter.
        **plotting_options : dict, default: None
            Keyword arguments. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.

        Returns
        -------
        List[Any]
            List with the picked bodies in the picked order.
        """
        self.add(
            object=object,
            merge_bodies=merge_bodies,
            merge_component=merge_component,
            filter=filter,
            **plotting_options,
        )

        if self._pl.geom_object_actors_map:
            self._geom_object_actors_map = self._pl.geom_object_actors_map
        else:
            logger.warning("No actors added to the plotter.")

        self.compute_edge_object_map()
        # Compute mapping between the objects and its edges.

        if view_2d is not None:
            self._pl.scene.view_vector(
                vector=view_2d["vector"],
                viewup=view_2d["viewup"],
            )

        # Enable widgets and picking capabilities
        self.enable_widgets()
        if self._allow_picking:
            self.enable_picking()

        # Update all buttons/widgets
        [widget.update() for widget in self._widgets]

        self.show_plotter(screenshot)

        picked_objects_list = []
        if isinstance(object, list):
            # Keep them ordered based on picking
            for name in self._picked_list:
                for elem in object:
                    if hasattr(elem, "name") and elem.name == name:
                        picked_objects_list.append(elem)
        elif hasattr(object, "name") and object.name in self._picked_list:
            picked_objects_list = [object]

        return picked_objects_list

    def show_plotter(self, screenshot: Optional[str] = None) -> None:
        """
        Show the plotter or start the `trame <https://kitware.github.io/trame/index.html>`_ service.

        Parameters
        ----------
        plotter : Plotter
            PyAnsys Geometry plotter with the meshes added.
        screenshot : str, default: None
            Path for saving a screenshot of the image that is being represented.
        """
        if self._use_trame and _HAS_TRAME:
            visualizer = TrameVisualizer()
            visualizer.set_scene(self._pl)
            visualizer.show()
        else:
            self._pl.show(screenshot=screenshot)

        pv.OFF_SCREEN = self._pv_off_screen_original



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing fundamental data of an assembly."""
from beartype.typing import TYPE_CHECKING, List

from ansys.geometry.core.designer.body import MasterBody
from ansys.geometry.core.math.constants import IDENTITY_MATRIX44
from ansys.geometry.core.math.matrix import Matrix44

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.component import Component


class Part:
    """
    Represents a part master.

    This class should not be accessed by users. The ``Part`` class holds fundamental
    data of an assembly.

    Parameters
    ----------
    id : str
        Unique identifier for the part.
    name : str
        Name of the part.
    components : List[MasterComponent]
        List of ``MasterComponent`` children that the part contains.
    bodies : List[MasterBody]
        List of ``MasterBody`` children that the part contains. These are master bodies.
    """

    def __init__(
        self, id: str, name: str, components: List["MasterComponent"], bodies: List[MasterBody]
    ) -> None:
        """Initialize the ``Part`` class."""
        self._id: str = id
        self._name: str = name
        self._components: List["MasterComponent"] = components
        self._bodies: List[MasterBody] = bodies

    @property
    def id(self) -> str:
        """ID of the part."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the part."""
        return self._name

    @property
    def components(self) -> List["MasterComponent"]:
        """``MasterComponent`` children that the part contains."""
        return self._components

    @components.setter
    def components(self, components: List["MasterComponent"]) -> None:
        self._components = components

    @property
    def bodies(self) -> List[MasterBody]:
        """
        ``MasterBody`` children that the part contains.

        These are master bodies.
        """
        return self._bodies

    @bodies.setter
    def bodies(self, bodies: List["MasterBody"]) -> None:
        self._bodies = bodies

    def __repr__(self) -> str:
        """Represent the part as a string."""
        return (
            f"Part(id={self.id}, "
            f"name={self.name}, "
            f"parts={[p.name for p in self.components]}, "
            f"bodies={[b.name for b in self.bodies]})"
        )


class MasterComponent:
    """
    Represents a part occurrence.

    Notes
    -----
    This class should not be accessed by users. It holds the fundamental data of
    an assembly. Master components wrap parts by adding a transform matrix.

    Parameters
    ----------
    id : str
        Unique identifier for the transformed part.
    name : str
        Name of the transformed part.
    part : Part
        Reference to the transformed part's master part.
    transform : Matrix44
        4x4 transformation matrix from the master part.
    """

    def __init__(
        self, id: str, name: str, part: Part, transform: Matrix44 = IDENTITY_MATRIX44
    ) -> None:
        """Initialize the ``MasterComponent`` class."""
        self._id: str = id
        self._name: str = name
        self._part: Part = part
        part.components.append(self)
        self._transform: Matrix44 = transform
        self._occurrences: List["Component"] = []

    @property
    def id(self) -> str:
        """ID of the transformed part."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the transformed part."""
        return self._name

    @property
    def occurrences(self) -> List["Component"]:
        """List of all occurrences of the component."""
        return self._occurrences

    @property
    def part(self) -> Part:
        """Master part of the transformed part."""
        return self._part

    @property
    def transform(self) -> Matrix44:
        """4x4 transformation matrix from the master part."""
        return self._transform

    @transform.setter
    def transform(self, matrix: Matrix44) -> None:
        self._transform = matrix

    def __repr__(self) -> str:
        """Represent the master component as a string."""
        return (
            f"MasterComponent(id={self.id}, "
            f"name={self.name}, "
            f"template={self.part}, "
            f"transform={self.transform})"
        )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a circle."""

from beartype import beartype as check_input_types
from beartype.typing import Optional, Union
from pint import Quantity
import pyvista as pv

from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point2D, Point3D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Distance
from ansys.geometry.core.shapes.curves.circle import Circle
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.typing import Real


class SketchCircle(SketchFace, Circle):
    """
    Provides for modeling a circle.

    Parameters
    ----------
    center: Point2D
        Center point of the circle.
    radius : Union[Quantity, Distance, Real]
        Radius of the circle.
    plane : Plane, optional
        Plane containing the sketched circle, which is the global XY plane
        by default.
    """

    @check_input_types
    def __init__(
        self, center: Point2D, radius: Union[Quantity, Distance, Real], plane: Plane = Plane()
    ):
        """Initialize the circle."""
        # Call SketchFace init method
        SketchFace.__init__(self)

        # Store the 2D center of the circle
        self._center = center

        self._radius = radius if isinstance(radius, Distance) else Distance(radius)
        if self._radius.value <= 0:
            raise ValueError("Radius must be a real positive value.")

        # Call Circle init method
        self._init_primitive_circle_from_plane(plane, radius=radius)

    def _init_primitive_circle_from_plane(
        self, plane: Plane, radius: Optional[Union[Quantity, Distance]] = None
    ) -> None:
        """
        Initialize correctly the underlying primitive ``Circle`` class.

        Parameters
        ----------
        plane : Plane
            Plane containing the sketched circle.
        radius : [Union[Quantity, Distance]], default: None
            Radius of the circle (if any).
        """
        # Use the radius given (if any)
        radius = radius if radius else self.radius

        # Call Circle init method
        center_global = plane.origin + Point3D(
            self.center[0] * plane.direction_x + self.center[1] * plane.direction_y,
            unit=self.center.base_unit,
        )
        Circle.__init__(self, center_global, radius, plane.direction_x, plane.direction_z)

    @property
    def center(self) -> Point2D:
        """Center of the circle."""
        return self._center

    @property
    def perimeter(self) -> Quantity:
        """
        Perimeter of the circle.

        Notes
        -----
        This property resolves the dilemma between using the ``SkethFace.perimeter``
        property and the ``Circle.perimeter`` property.
        """
        return Circle.perimeter.fget(self)

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """
        VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        circle = pv.Circle(self.radius.m_as(DEFAULT_UNITS.LENGTH))
        return circle.translate(
            [
                self.center.x.m_as(DEFAULT_UNITS.LENGTH),
                self.center.y.m_as(DEFAULT_UNITS.LENGTH),
                0,
            ],
            inplace=True,
        )

    def plane_change(self, plane: Plane) -> None:
        """
        Redefine the plane containing the ``SketchCircle`` objects.

        Notes
        -----
        This implies that their 3D definition might suffer changes.

        Parameters
        ----------
        plane : Plane
            Desired new plane that is to contain the sketched circle.
        """
        # Reinitialize the circle definition for the given plane
        self._init_primitive_circle_from_plane(plane)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for connecting to a local Docker container with the Geometry service."""
from enum import Enum
from functools import wraps
import os

from beartype import beartype as check_input_types
from beartype.typing import Optional, Tuple, Union

try:
    from docker.client import DockerClient
    from docker.errors import APIError, ContainerError, ImageNotFound
    from docker.models.containers import Container

    _HAS_DOCKER = True
except ModuleNotFoundError:  # pragma: no cover
    _HAS_DOCKER = False

from ansys.geometry.core.connection.defaults import DEFAULT_PORT, GEOMETRY_SERVICE_DOCKER_IMAGE
from ansys.geometry.core.logger import LOG as logger


def _docker_python_available(func):
    """
    Check whether Docker is installed as a Python package.

    This function works as a decorator.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not _HAS_DOCKER:  # pragma: no cover
            raise ModuleNotFoundError(
                "The package 'docker' is required to use the LocalDockerInstance class."
            )
        else:
            return func(*args, **kwargs)

    return wrapper


class GeometryContainers(Enum):
    """Provides an enum holding the available Geometry services."""

    WINDOWS_LATEST = 0, "windows", "windows-latest"
    LINUX_LATEST = 1, "linux", "linux-latest"
    WINDOWS_LATEST_UNSTABLE = 2, "windows", "windows-latest-unstable"
    LINUX_LATEST_UNSTABLE = 3, "linux", "linux-latest-unstable"
    WINDOWS_24_1 = 4, "windows", "windows-24.1"
    LINUX_24_1 = 5, "linux", "linux-24.1"
    WINDOWS_24_2 = 6, "windows", "windows-24.2"
    LINUX_24_2 = 7, "linux", "linux-24.2"


class LocalDockerInstance:
    """
    Instantiates a Geometry service as a local Docker container.

    By default, if a container with the Geometry service already exists at the given port,
    PyAnsys Geometry connects to it. Otherwise, PyAnsys Geometry tries to launch its own service.

    Parameters
    ----------
    port : int, optional
        Localhost port to deploy the Geometry service on or the
        the ``Modeler`` interface to connect to (if it is already deployed). By default,
        the value is the one for the ``DEFAULT_PORT`` connection parameter.
    connect_to_existing_service : bool, default: True
        Whether the ``Modeler`` interface should connect to a Geometry
        service already deployed at the specified port.
    restart_if_existing_service : bool, default: False
        Whether the Geometry service (which is already running)
        should be restarted when attempting connection.
    name : Optional[str], default: None
        Name of the Docker container to deploy. The default is ``None``,
        in which case Docker assigns it a random name.
    image : Optional[GeometryContainers], default: None
        The Geometry service Docker image to deploy. The default is ``None``,
        in which case the ``LocalDockerInstance`` class identifies the OS of your
        Docker engine and deploys the latest version of the Geometry service for that
        OS.
    """

    __DOCKER_CLIENT__: "DockerClient" = None
    """
    Docker client class variable. The default is ``None``, in which case lazy
    initialization is used.

    Notes
    -----
    ``__DOCKER_CLIENT__`` is a class variable, meaning that it is
    the same variable for all instances of this class.
    """

    @staticmethod
    @_docker_python_available
    def docker_client() -> "DockerClient":
        """
        Get the initialized ``__DOCKER_CLIENT__`` object.

        Notes
        -----
        The ``LocalDockerInstance`` class performs a lazy initialization of the
        ``__DOCKER_CLIENT__`` class variable.

        Returns
        -------
        ~docker.client.DockerClient
            Initialized Docker client.
        """
        if not LocalDockerInstance.__DOCKER_CLIENT__:
            LocalDockerInstance.__DOCKER_CLIENT__ = DockerClient.from_env()

        return LocalDockerInstance.__DOCKER_CLIENT__

    @staticmethod
    @_docker_python_available
    def is_docker_installed() -> bool:
        """
        Check whether a local installation of Docker engine is available and running.

        Returns
        -------
        bool
            ``True`` if Docker engine is available and running, ``False`` otherwise.
        """
        # When trying to ping, this throws an error if the
        # server is not available
        try:
            return LocalDockerInstance.docker_client().ping()
        except APIError:  # pragma: no cover
            return False

    @check_input_types
    @_docker_python_available
    def __init__(
        self,
        port: int = DEFAULT_PORT,
        connect_to_existing_service: bool = True,
        restart_if_existing_service: bool = False,
        name: Optional[str] = None,
        image: Optional[GeometryContainers] = None,
    ) -> None:
        """``LocalDockerInstance`` constructor."""
        # Initialize instance variables
        self._container: Container = None
        self._existed_previously: bool = False

        # Check the port availability
        port_available, cont = self._check_port_availability(port)

        # If a service was already deployed... check if it is a Geometry service
        # in case you want to connect to it
        if cont and connect_to_existing_service and self._is_cont_geom_service(cont):
            # Now, check if a restart of the service is required
            if restart_if_existing_service:
                logger.warning(f"Restarting service already running at port {port}...")
                cont.restart()

            # Finally, store the container
            self._container = cont
            self._existed_previously = True
            return

        # At this stage, confirm that you have to deploy our own Geometry service.
        #
        # First, check if the port is available... otherwise raise error
        if port_available:
            self._deploy_container(port=port, name=name, image=image)
        else:
            raise RuntimeError(f"Geometry service cannot be deployed on port {port}")

    def _check_port_availability(self, port: int) -> Tuple[bool, Optional["Container"]]:
        """
        Check whether the requested port is available for deployment.

        Returns
        -------
        tuple[bool, Container]
            Tuple with a Boolean indicating whether the port is available.
            If the port is not available, the container deployed at that port is also returned.
        """
        # First, check if there is a container already running at that port
        for cont in self.docker_client().containers.list():
            for _, ports_shared in cont.attrs["NetworkSettings"]["Ports"].items():
                # If no ports are shared, continue looping
                if not ports_shared:
                    continue

                # Check shared ports otherwise
                for port_shared in ports_shared:
                    if int(port_shared["HostPort"]) == port:
                        logger.warning(f"Service is already running at port {port}...")
                        return (False, cont)

        # If you reached here, just return default values
        return (True, None)

    def _is_cont_geom_service(self, cont: "Container") -> bool:
        """
        Check whether a provided ``Container`` object is a Geometry service container.

        Parameters
        ----------
        cont : Container
            Container deployed to check.

        Returns
        -------
        bool
            Returns ``True`` if the container provided is a Geometry service.
            Otherwise, ``False``.
        """
        # If one of the tags matches a Geometry service tag --> Return True
        for tag in cont.image.tags:
            for geom_services in GeometryContainers:
                if tag == f"{GEOMETRY_SERVICE_DOCKER_IMAGE}:{geom_services.value[2]}":
                    return True

        # If you have reached this point, the image is not a Geometry service
        return False  # pragma: no cover

    def _deploy_container(
        self, port: int, name: Union[str, None], image: Union[GeometryContainers, None]
    ):
        """
        Handle the deployment of a Geometry service according to the arguments.

        Parameters
        ----------
        port : int
            Port under which the service should be deployed.
        name : Union[str, None], optional
            Name given to the deployed container. If ``None``, Docker will provide
            an arbitrary name.
        image : Union[GeometryContainers, None]
            Geometry service Docker container image to be used. If ``None``, the
            latest container version matching

        Raises
        ------
        RuntimeError
            In case the Geometry service cannot be launched.
        """
        # First get the Docker Engine OS
        docker_os = self.docker_client().info()["OSType"]

        # If no image is provided, default to the whatever Docker engine OS
        # your system is running on... Latest images are used because they are the
        # ones defined first in the GeometryContainers enum class.
        if image is None:
            for geom_service in GeometryContainers:
                if geom_service.value[1] == docker_os:
                    image = geom_service
                    break

        # If image is still None, this means it cannot be deployed on your OS
        #
        # Also, if you requested an image incompatible with the existing OS, it
        # is not possible either.
        if image is None or image.value[1] != docker_os:  # pragma: no cover
            raise RuntimeError(f"Geometry service cannot be launched on {docker_os}")

        # At this point, you are can deploy the Geometry service.
        #
        # Check if the license server env variable is available
        license_server = os.getenv("ANSRV_GEO_LICENSE_SERVER", None)
        if not license_server:  # pragma: no cover
            raise RuntimeError(
                f"No license server provided... Store its value under the following env variable: ANSRV_GEO_LICENSE_SERVER."  # noqa: E501
            )

        # Try to deploy it
        try:
            container: Container = self.docker_client().containers.run(
                image=f"{GEOMETRY_SERVICE_DOCKER_IMAGE}:{image.value[2]}",
                detach=True,
                auto_remove=True,
                name=name,
                ports={"50051/tcp": port},
                environment={
                    "LICENSE_SERVER": license_server,
                    "LOG_LEVEL": os.getenv("ANSRV_GEO_LOG_LEVEL", 2),
                    "ENABLE_TRACE": os.getenv("ANSRV_GEO_ENABLE_TRACE", 0),
                    "USE_DEBUG_MODE": os.getenv("ANSRV_GEO_USE_DEBUG_MODE", 0),
                },
            )
        except ImageNotFound as err:  # pragma: no cover
            raise RuntimeError(
                f"Geometry service Docker image {image.value[1]} not found. Download it first to your machine."  # noqa: E501
            )
        except (ContainerError, APIError) as err:
            raise RuntimeError(
                f"Geometry service Docker image error when initialized. Error: {err.explanation}"
            )

        # If the deployment went fine, this means that you have deployed the service.
        self._container = container
        self._existed_previously = False

    @property
    def container(self) -> "Container":
        """Docker container object that hosts the deployed Geometry service."""
        return self._container

    @property
    def existed_previously(self) -> bool:
        """
        Flag indicating whether the container previously existed.

        Returns ``False`` if the Geometry service was effectively
        deployed by this class or ``True`` if it already existed.
        """
        return self._existed_previously


def get_geometry_container_type(instance: LocalDockerInstance) -> Union[GeometryContainers, None]:
    """
    Given a ``LocalDockerInstance``, provide back the ``GeometryContainers`` value.

    Notes
    -----
    This method returns the first hit on the available tags.

    Parameters
    ----------
    instance : LocalDockerInstance
        The LocalDockerInstance object.

    Returns
    -------
    Union[GeometryContainers, None]
        The GeometryContainer value corresponding to the previous image or None
        if not match.
    """
    for tag in instance.container.image.tags:
        for geom_services in GeometryContainers:
            if tag == f"{GEOMETRY_SERVICE_DOCKER_IMAGE}:{geom_services.value[2]}":
                return geom_services

    return None



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for using `trame <https://kitware.github.io/trame/index.html>`_ for visualization."""
from beartype.typing import TYPE_CHECKING, Union
import pyvista as pv

from ansys.geometry.core import LOG as logger
from ansys.geometry.core.plotting.plotter import Plotter

if TYPE_CHECKING:
    from ansys.geometry.core.plotting.plotter_helper import PlotterHelper

try:
    from pyvista.trame.ui import plotter_ui
    from trame.app import get_server
    from trame.ui import vuetify, vuetify3

    _HAS_TRAME = True

except ModuleNotFoundError:  # pragma: no cover
    _HAS_TRAME = False


class TrameVisualizer:
    """
    Defines the trame layout view.

    Parameters
    ----------
    client_type : str, optional
        Client type to use for the trame server. Options are 'vue2' and 'vue3'.
        Default is 'vue3'.
    """

    def __init__(self, client_type: str = "vue3") -> None:
        """Initialize the trame server and server-related variables."""
        if not _HAS_TRAME:  # pragma: no cover
            raise ModuleNotFoundError(
                "The package 'pyvista[trame]' is required to use this function."
            )
        self._client_type = client_type
        if client_type not in ["vue2", "vue3"]:
            self._client_type = "vue3"
            logger.warning("Invalid client type {}. Defaulting to 'vue3'.", client_type)

        self.server = get_server(client_type=client_type)
        self.state, self.ctrl = self.server.state, self.server.controller

    def set_scene(self, plotter: Union["pv.Plotter", Plotter, "PlotterHelper"]):
        """
        Set the trame layout view and the mesh to show through the PyVista plotter.

        Parameters
        ----------
        plotter : Union[~pyvista.Plotter, ~Plotter, ~PlotterHelper]
            PyVista plotter or PyAnsys plotter to render.
        """
        from ansys.geometry.core.plotting.plotter_helper import PlotterHelper

        if isinstance(plotter, PlotterHelper):
            pv_plotter = plotter._pl.scene
        elif isinstance(plotter, Plotter):
            pv_plotter = plotter.scene
        elif isinstance(plotter, pv.Plotter):
            pv_plotter = plotter
        else:
            logger.warning("Invalid plotter type. Expected PyVista plotter or PyAnsys plotter.")

        self.state.trame__title = "PyAnsys Geometry Viewer"
        if self._client_type == "vue3":
            page_layout = vuetify3.SinglePageLayout(self.server)
        else:
            page_layout = vuetify.SinglePageLayout(self.server)

        with page_layout as layout:
            layout.icon.click = self.ctrl.view_reset_camera
            layout.title.set_text("PyAnsys Geometry")

            with layout.content:
                # Use PyVista UI template for Plotters
                view = plotter_ui(pv_plotter)
                self.ctrl.view_update = view.update

            # hide footer with trame watermark
            layout.footer.hide()

    def show(self):
        """Start the trame server and show the mesh."""
        self.server.start()



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Data types for plotting."""
from beartype.typing import Any, List
import pyvista as pv

from ansys.geometry.core.designer.edge import Edge
from ansys.geometry.core.logger import LOG


class EdgePlot:
    """
    Mapper class to relate PyAnsys Geometry edges with its PyVista actor.

    Parameters
    ----------
    actor : ~pyvista.Actor
        PyVista actor that represents the edge.
    edge_object : Edge
        PyAnsys Geometry edge that is represented by the PyVista actor.
    parent : GeomObjectPlot, optional
        Parent PyAnsys Geometry object of this edge, by default ``None``.
    """

    def __init__(self, actor: pv.Actor, edge_object: Edge, parent: "GeomObjectPlot" = None) -> None:
        """Initialize EdgePlot variables."""
        self._actor = actor
        self._object = edge_object
        self._parent = parent

    @property
    def actor(self) -> pv.Actor:
        """
        Return PyVista actor of the object.

        Returns
        -------
        ~pyvista.Actor
            PyVista actor.
        """
        return self._actor

    @property
    def edge_object(self) -> Edge:
        """
        Return the PyAnsys Geometry edge.

        Returns
        -------
        Edge
            PyAnsys Geometry edge.
        """
        return self._object

    @property
    def parent(self) -> Any:
        """
        Parent PyAnsys Geometry object of this edge.

        Returns
        -------
        Any
            PyAnsys Geometry object.
        """
        return self._parent

    @property
    def name(self) -> str:
        """
        Return the name of the edge.

        Returns
        -------
        str
            Name of the edge.
        """
        if self.parent:
            return f"{self.parent.name}-{self.edge_object.id}"
        else:
            return self.edge_object.id

    @parent.setter
    def parent(self, parent: "GeomObjectPlot"):
        """
        Set the parent object of the edge.

        Parameters
        ----------
        parent : GeomObjectPlot
            Parent of the edge.
        """
        self._parent = parent


class GeomObjectPlot:
    """
    Mapper class to relate PyAnsys Geometry objects with its PyVista actor.

    Parameters
    ----------
    actor : pv.Actor
        PyVista actor that represents the pyansys-geometry object.
    object : Any
        PyAnsys Geometry object that is represented.
    edges : List[EdgePlot], optional
        List of edges of the PyAnsys Geometry object, by default ``None``.
    add_body_edges: bool, optional
        Flag to specify if you want to be able to add edges.
    """

    def __init__(
        self,
        actor: pv.Actor,
        object: Any,
        edges: List[EdgePlot] = None,
        add_body_edges: bool = True,
    ) -> None:
        """Initialize GeomObjectPlot variables."""
        self._actor = actor
        self._object = object
        self._edges = edges
        self._add_body_edges = add_body_edges

    @property
    def actor(self) -> pv.Actor:
        """
        Return the PyVista actor of the PyAnsys Geometry object.

        Returns
        -------
        ~pyvista.Actor
            Actor of the PyAnsys Geometry object.
        """
        return self._actor

    @property
    def object(self) -> Any:
        """
        Return the PyAnsys Geometry object.

        Returns
        -------
        Any
            PyAnsys Geometry object.
        """
        return self._object

    @property
    def edges(self) -> List[EdgePlot]:
        """
        Return the list of edges associated to this PyAnsys Geometry object.

        Returns
        -------
        List[EdgePlot]
            List of the edges of this object.
        """
        return self._edges

    @edges.setter
    def edges(self, edges: List[EdgePlot]):
        """
        Set the edges of this object.

        Parameters
        ----------
        edges : List[EdgePlot]
            List of the edges of this object.
        """
        if self._add_body_edges:
            self._edges = edges
        else:
            LOG.warning(
                "To add edges to this body, it should be initialized with add_body_edges flag."
            )

    @property
    def name(self) -> str:
        """
        Return the name of this object.

        Returns
        -------
        str
            Name of the object.
        """
        return self._object.name

    @property
    def add_body_edges(self) -> bool:
        """
        Return whether you want to be able to add edges.

        Returns
        -------
        bool
            Flag to add edges.
        """
        return self._add_body_edges



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing an ellipse."""

from beartype import beartype as check_input_types
from beartype.typing import Optional, Union
import numpy as np
from pint import Quantity
import pyvista as pv
from scipy.spatial.transform import Rotation as spatial_rotation

from ansys.geometry.core.math.matrix import Matrix33, Matrix44
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point2D, Point3D
from ansys.geometry.core.math.vector import Vector3D
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.units import UNITS
from ansys.geometry.core.shapes.curves.ellipse import Ellipse
from ansys.geometry.core.sketch.face import SketchFace
from ansys.geometry.core.typing import Real


class SketchEllipse(SketchFace, Ellipse):
    """
    Provides for modeling an ellipse.

    Parameters
    ----------
    center: Point2D
        Center point of the ellipse.
    major_radius : Union[Quantity, Distance, Real]
        Major radius of the ellipse.
    minor_radius : Union[Quantity, Distance, Real]
        Minor radius of the ellipse.
    angle : Union[Quantity, Angle, Real], default: 0
        Placement angle for orientation alignment.
    plane : Plane, optional
        Plane containing the sketched ellipse, which is the global XY plane
        by default.
    """

    @check_input_types
    def __init__(
        self,
        center: Point2D,
        major_radius: Union[Quantity, Distance, Real],
        minor_radius: Union[Quantity, Distance, Real],
        angle: Optional[Union[Quantity, Angle, Real]] = 0,
        plane: Plane = Plane(),
    ):
        """Initialize the ellipse."""
        # Call SketchFace init method
        SketchFace.__init__(self)

        # Store the 2D center of the ellipse
        self._center = center

        self._major_radius = (
            major_radius if isinstance(major_radius, Distance) else Distance(major_radius)
        )
        self._minor_radius = (
            minor_radius if isinstance(minor_radius, Distance) else Distance(minor_radius)
        )

        self._angle_offset = angle if isinstance(angle, Angle) else Angle(angle)

        # Call Ellipse init method
        self._init_primitive_ellipse_from_plane(
            plane, self._major_radius, self._minor_radius, self._angle_offset
        )

    def _init_primitive_ellipse_from_plane(
        self,
        plane: Plane,
        major_radius: Optional[Distance] = None,
        minor_radius: Optional[Distance] = None,
        angle: Optional[Angle] = None,
    ) -> None:
        """
        Initialize correctly the underlying primitive ``Ellipse`` class.

        Parameters
        ----------
        plane : Plane
            Plane containing the sketched ellipse.
        major_radius : [Distance], default: None
            Major radius of the ellipse (if any).
        minor_radius : [Distance], default: None
            Minor radius of the ellipse (if any).
        angle : [Angle], default: None
            Placement angle for orientation alignment.
        """
        major_radius = major_radius if major_radius else self.major_radius
        minor_radius = minor_radius if minor_radius else self.minor_radius
        angle = angle if angle else self.angle

        # Call Ellipse init method
        center_global = plane.origin + Point3D(
            self.center[0] * plane.direction_x + self.center[1] * plane.direction_y,
            unit=self.center.base_unit,
        )

        angle_rad = angle.value.m_as(UNITS.radian)
        new_rotated_dir_x = Vector3D(
            [
                np.cos(angle_rad) * plane.direction_x.x - np.sin(angle_rad) * plane.direction_x.y,
                np.sin(angle_rad) * plane.direction_x.x + np.cos(angle_rad) * plane.direction_x.y,
                plane.direction_x.z,
            ]
        )

        Ellipse.__init__(
            self, center_global, major_radius, minor_radius, new_rotated_dir_x, plane.direction_z
        )

    @property
    def center(self) -> Point2D:
        """Center point of the ellipse."""
        return self._center

    @property
    def angle(self) -> Quantity:
        """Orientation angle of the ellipse."""
        return self._angle_offset.value

    @property
    def perimeter(self) -> Quantity:
        """
        Perimeter of the circle.

        Notes
        -----
        This property resolves the dilemma between using the ``SkethFace.perimeter``
        property and the ``Ellipse.perimeter`` property.
        """
        return Ellipse.perimeter.fget(self)

    @property
    def visualization_polydata(self) -> pv.PolyData:
        """
        VTK polydata representation for PyVista visualization.

        The representation lies in the X/Y plane within
        the standard global Cartesian coordinate system.

        Returns
        -------
        pyvista.PolyData
            VTK pyvista.Polydata configuration.
        """
        rotation = Matrix33(
            spatial_rotation.from_euler(
                "xyz", [0, 0, self._angle_offset.value.m_as(UNITS.radian)], degrees=False
            ).as_matrix()
        )

        transformation_matrix = Matrix44(
            [
                [
                    rotation[0, 0],
                    rotation[0, 1],
                    rotation[0, 2],
                    self.center.x.m_as(DEFAULT_UNITS.LENGTH),
                ],
                [
                    rotation[1, 0],
                    rotation[1, 1],
                    rotation[1, 2],
                    self.center.y.m_as(DEFAULT_UNITS.LENGTH),
                ],
                [
                    rotation[2, 0],
                    rotation[2, 1],
                    rotation[2, 2],
                    0,
                ],
                [0, 0, 0, 1],
            ]
        )

        return pv.Ellipse(
            semi_major_axis=self.major_radius.m_as(DEFAULT_UNITS.LENGTH),
            semi_minor_axis=self.minor_radius.m_as(DEFAULT_UNITS.LENGTH),
        ).transform(transformation_matrix)

    def plane_change(self, plane: Plane) -> None:
        """
        Redefine the plane containing ``SketchEllipse`` objects.

        Notes
        -----
        This implies that their 3D definition might suffer changes.

        Parameters
        ----------
        plane : Plane
            Desired new plane that is to contain the sketched ellipse.
        """
        # Reinitialize the Circle definition for the given plane
        self._init_primitive_ellipse_from_plane(plane)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides tools for repairing bodies."""

from ansys.api.geometry.v0.repairtools_pb2 import (
    FindDuplicateFacesRequest,
    FindExtraEdgesRequest,
    FindInexactEdgesRequest,
    FindMissingFacesRequest,
    FindSmallFacesRequest,
    FindSplitEdgesRequest,
    FindStitchFacesRequest,
)
from ansys.api.geometry.v0.repairtools_pb2_grpc import RepairToolsStub
from beartype.typing import TYPE_CHECKING, List
from google.protobuf.wrappers_pb2 import DoubleValue

from ansys.geometry.core.connection import GrpcClient
from ansys.geometry.core.misc.auxiliary import (
    get_bodies_from_ids,
    get_design_from_body,
    get_edges_from_ids,
    get_faces_from_ids,
)
from ansys.geometry.core.tools.problem_areas import (
    DuplicateFaceProblemAreas,
    ExtraEdgeProblemAreas,
    InexactEdgeProblemAreas,
    MissingFaceProblemAreas,
    SmallFaceProblemAreas,
    SplitEdgeProblemAreas,
    StitchFaceProblemAreas,
)
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body


class RepairTools:
    """Repair tools for PyAnsys Geometry."""

    def __init__(self, grpc_client: GrpcClient):
        """Initialize Repair Tools class."""
        self._grpc_client = grpc_client
        self._repair_stub = RepairToolsStub(self._grpc_client.channel)

    def find_split_edges(
        self, bodies: List["Body"], angle: Real = 0.0, length: Real = 0.0
    ) -> List[SplitEdgeProblemAreas]:
        """
        Find split edges in the given list of bodies.

        This method finds the split edge problem areas and returns a list of split edge
        problem areas objects.

        Parameters
        ----------
        bodies : List[Body]
            List of bodies that split edges are investigated on.
        angle : Real
            The maximum angle between edges.
        length : Real
            The maximum length of the edges.

        Returns
        -------
        List[SplitEdgeProblemAreas]
            List of objects representing split edge problem areas.
        """
        if not bodies:
            return []

        angle_value = DoubleValue(value=float(angle))
        length_value = DoubleValue(value=float(length))
        body_ids = [body.id for body in bodies]

        problem_areas_response = self._repair_stub.FindSplitEdges(
            FindSplitEdgesRequest(
                bodies_or_faces=body_ids, angle=angle_value, distance=length_value
            )
        )
        parent_design = get_design_from_body(bodies[0])
        return [
            SplitEdgeProblemAreas(
                str(res.id),
                self._grpc_client,
                get_edges_from_ids(parent_design, res.edge_monikers),
            )
            for res in problem_areas_response.result
        ]

    def find_extra_edges(self, bodies: List["Body"]) -> List[ExtraEdgeProblemAreas]:
        """
        Find the extra edges in the given list of bodies.

        This method find the extra edge problem areas and returns a list of extra edge
        problem areas objects.

        Parameters
        ----------
        bodies : List[Body]
            List of bodies that extra edges are investigated on.

        Returns
        -------
        List[ExtraEdgeProblemArea]
            List of objects representing extra edge problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]
        problem_areas_response = self._repair_stub.FindExtraEdges(
            FindExtraEdgesRequest(selection=body_ids)
        )
        parent_design = get_design_from_body(bodies[0])

        return [
            ExtraEdgeProblemAreas(
                str(res.id),
                self._grpc_client,
                get_edges_from_ids(parent_design, res.edge_monikers),
            )
            for res in problem_areas_response.result
        ]

    def find_inexact_edges(self, bodies: List["Body"]) -> List[InexactEdgeProblemAreas]:
        """
        Find inexact edges in the given list of bodies.

        This method find the inexact edge problem areas and returns a list of inexact
        edge problem areas objects.

        Parameters
        ----------
        bodies : List[Body]
            List of bodies that inexact edges are investigated on.

        Returns
        -------
        List[InExactEdgeProblemArea]
            List of objects representing inexact edge problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]
        problem_areas_response = self._repair_stub.FindInexactEdges(
            FindInexactEdgesRequest(selection=body_ids)
        )

        parent_design = get_design_from_body(bodies[0])

        return [
            InexactEdgeProblemAreas(
                str(res.id),
                self._grpc_client,
                get_edges_from_ids(parent_design, res.edge_monikers),
            )
            for res in problem_areas_response.result
        ]

    def find_duplicate_faces(self, bodies: List["Body"]) -> List[DuplicateFaceProblemAreas]:
        """
        Find the duplicate face problem areas.

        This method finds the duplicate face problem areas and returns a list of
        duplicate face problem areas objects.

        Parameters
        ----------
        bodies : List[Body]
            List of bodies that duplicate faces are investigated on.

        Returns
        -------
        List[DuplicateFaceProblemAreas]
            List of objects representing duplicate face problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]
        problem_areas_response = self._repair_stub.FindDuplicateFaces(
            FindDuplicateFacesRequest(faces=body_ids)
        )

        parent_design = get_design_from_body(bodies[0])
        return [
            DuplicateFaceProblemAreas(
                str(res.id),
                self._grpc_client,
                get_faces_from_ids(parent_design, res.face_monikers),
            )
            for res in problem_areas_response.result
        ]

    def find_missing_faces(self, bodies: List["Body"]) -> List[MissingFaceProblemAreas]:
        """
        Find the missing faces.

        This method find the missing face problem areas and returns a list of missing
        face problem areas objects.

        Parameters
        ----------
        bodies : List[Body]
            List of bodies that missing faces are investigated on.

        Returns
        -------
        List[MissingFaceProblemAreas]
            List of objects representing missing face problem areas.
        """
        if not bodies:
            return []
        body_ids = [body.id for body in bodies]
        problem_areas_response = self._repair_stub.FindMissingFaces(
            FindMissingFacesRequest(faces=body_ids)
        )
        parent_design = get_design_from_body(bodies[0])

        return [
            MissingFaceProblemAreas(
                str(res.id),
                self._grpc_client,
                get_edges_from_ids(parent_design, res.edge_monikers),
            )
            for res in problem_areas_response.result
        ]

    def find_small_faces(self, bodies: List["Body"]) -> List[SmallFaceProblemAreas]:
        """
        Find the small face problem areas.

        This method finds and returns a list of ids of small face problem areas
        objects.

        Parameters
        ----------
        bodies : List[Body]
            List of bodies that small faces are investigated on.

        Returns
        -------
        List[SmallFaceProblemAreas]
            List of objects representing small face problem areas.
        """
        if not bodies:
            return []

        body_ids = [body.id for body in bodies]
        problem_areas_response = self._repair_stub.FindSmallFaces(
            FindSmallFacesRequest(selection=body_ids)
        )
        parent_design = get_design_from_body(bodies[0])

        return [
            SmallFaceProblemAreas(
                str(res.id),
                self._grpc_client,
                get_faces_from_ids(parent_design, res.face_monikers),
            )
            for res in problem_areas_response.result
        ]

    def find_stitch_faces(self, bodies: List["Body"]) -> List[StitchFaceProblemAreas]:
        """
        Return the list of stitch face problem areas.

        This method find the stitch face problem areas and returns a list of ids of stitch face
        problem areas objects.

        Parameters
        ----------
        bodies : List[Body]
            List of bodies that stitchable faces are investigated on.

        Returns
        -------
        List[StitchFaceProblemAreas]
            List of objects representing stitch face problem areas.
        """
        body_ids = [body.id for body in bodies]
        problem_areas_response = self._repair_stub.FindStitchFaces(
            FindStitchFacesRequest(faces=body_ids)
        )
        parent_design = get_design_from_body(bodies[0])
        return [
            StitchFaceProblemAreas(
                str(res.id),
                self._grpc_client,
                get_bodies_from_ids(parent_design, res.body_monikers),
            )
            for res in problem_areas_response.result
        ]



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the view button widget for changing the camera view."""

from enum import Enum

from pyvista import Plotter

from ansys.geometry.core.plotting.widgets.button import Button


class ViewDirection(Enum):
    """Provides an enum with the available views."""

    XYPLUS = 0, "+xy.png", (5, 220)
    XYMINUS = 1, "-xy.png", (5, 251)
    XZPLUS = 2, "+xz.png", (5, 282)
    XZMINUS = 3, "-xz.png", (5, 313)
    YZPLUS = 4, "+yz.png", (5, 344)
    YZMINUS = 5, "-yz.png", (5, 375)
    ISOMETRIC = 6, "isometric.png", (5, 406)


class ViewButton(Button):
    """
    Provides for changing the view.

    Parameters
    ----------
    plotter : Plotter
        Plotter to draw the buttons on.
    direction : ViewDirection
        Direction of the view.
    """

    def __init__(self, plotter: Plotter, direction: tuple):
        """Initialize the ``ViewButton`` class."""
        super().__init__(plotter, direction)
        self.direction = direction

    def callback(self, state: bool) -> None:
        """
        Change the view depending on button interaction.

        Parameters
        ----------
        state : bool
            State of the button, which is inherited from PyVista. The value is ``True``
            if the button is active.

        Raises
        ------
        NotImplementedError
            Raised if the specified direction is not implemented.
        """
        if self.direction == ViewDirection.XYPLUS:
            self.plotter.view_xy()
        elif self.direction == ViewDirection.XYMINUS:
            self.plotter.view_yx()
        elif self.direction == ViewDirection.XZPLUS:
            self.plotter.view_xz()
        elif self.direction == ViewDirection.XZMINUS:
            self.plotter.view_zx()
        elif self.direction == ViewDirection.YZPLUS:
            self.plotter.view_yz()
        elif self.direction == ViewDirection.YZMINUS:
            self.plotter.view_zy()
        elif self.direction == ViewDirection.ISOMETRIC:
            self.plotter.view_isometric()
        else:  # pragma: no cover
            raise NotImplementedError(
                f"ViewDirection {self.direction.name} is not implemented as a widget."
            )



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Module to perform a connection validation check.

The method in this module is only used for testing the default Docker service on
GitHub and can safely be skipped within testing.

This command shows how this method is typically used:

.. code:: bash

   python -c "from ansys.geometry.core.connection import validate; validate()"
"""
from ansys.geometry.core.connection.client import GrpcClient


def validate(*args, **kwargs):  # pragma: no cover
    """Create a client using the default settings and validate it."""
    print(GrpcClient(*args, **kwargs))
    # TODO: consider adding additional server stat reporting



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides tools for measurement."""

from ansys.api.geometry.v0.measuretools_pb2 import (
    MinDistanceBetweenObjectsRequest,
    MinDistanceBetweenObjectsResponse,
)
from ansys.api.geometry.v0.measuretools_pb2_grpc import MeasureToolsStub
from beartype.typing import TYPE_CHECKING

from ansys.geometry.core.connection import GrpcClient
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.misc.checks import min_backend_version
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Distance

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body


class Gap:
    """
    Represents a gap between two bodies.

    Parameters
    ----------
    distance : Distance
        Distance between two sides of the gap.
    """

    def __init__(self, distance: Distance):
        """Initialize ``Gap`` class."""
        self._distance = distance

    @property
    def distance(self) -> Distance:
        """Returns the closest distance between two bodies."""
        return self._distance

    @classmethod
    def _from_distance_response(cls, response: MinDistanceBetweenObjectsResponse) -> "Gap":
        """
        Construct ``Gap`` object from distance response.

        Notes
        -----
        This method is used internally to construct a ``Gap`` object from a
        gRPC response.

        Parameters
        ----------
        response : MinDistanceBetweenObjectsResponse
            Response from the gRPC server.
        """
        distance = Distance(response.gap.distance, unit=DEFAULT_UNITS.LENGTH)
        return cls(distance)


class MeasurementTools:
    """
    Measurement tools for PyAnsys Geometry.

    Parameters
    ----------
    grpc_client : GrpcClient
        gRPC client to use for the measurement tools.
    """

    @protect_grpc
    def __init__(self, grpc_client: GrpcClient):
        """Initialize measurement tools class."""
        self._grpc_client = grpc_client
        self._measure_stub = MeasureToolsStub(self._grpc_client.channel)

    @protect_grpc
    @min_backend_version(24, 2, 0)
    def min_distance_between_objects(self, body1: "Body", body2: "Body") -> Gap:
        """
        Find the gap between two bodies.

        Parameters
        ----------
        body1 : Body
            First body to measure the gap.
        body2 : Body
            Second body to measure the gap.

        Returns
        -------
        Gap
            Gap between two bodies.
        """
        response = self._measure_stub.MinDistanceBetweenObjects(
            MinDistanceBetweenObjectsRequest(bodies=[body1.id, body2.id])
        )
        return Gap._from_distance_response(response)



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the parametrization-related classes."""

from enum import Enum, unique

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.misc.accuracy import Accuracy
from ansys.geometry.core.typing import Real


class ParamUV:
    """
    Parameter class containing 2 parameters: (u, v).

    Notes
    -----
    Likened to a 2D point in UV space Used as an argument in parametric
    surface evaluations. This matches the service implementation for the
    Geometry service.

    Parameters
    ----------
    u : Real
        u-parameter.
    v : Real
        v-parameter.
    """

    def __init__(self, u: Real, v: Real) -> None:
        """Initialize ``ParamUV`` class."""
        self._u = u
        self._v = v

    @property
    def u(self) -> Real:
        """u-parameter."""
        return self._u

    @property
    def v(self) -> Real:
        """v-parameter."""
        return self._v

    @check_input_types
    def __add__(self, other: "ParamUV") -> "ParamUV":
        """
        Add the u and v components of the other ParamUV to this ParamUV.

        Parameters
        ----------
        other : ParamUV
            The parameters to add these parameters.

        Returns
        -------
        ParamUV
            The sum of the parameters.
        """
        return ParamUV(self._u + other._u, self._v + other._v)

    @check_input_types
    def __sub__(self, other: "ParamUV") -> "ParamUV":
        """
        Subtract the u and v components of the other ParamUV from this ParamUV.

        Parameters
        ----------
        other : ParamUV
            The parameters to subtract from these parameters.

        Returns
        -------
        ParamUV
            The difference of the parameters.
        """
        return ParamUV(self._u - other._u, self._v - other._v)

    @check_input_types
    def __mul__(self, other: "ParamUV") -> "ParamUV":
        """
        Multiplies the u and v components of this ParamUV by the other ParamUV.

        Parameters
        ----------
        other : ParamUV
            The parameters to multiply by these parameters.

        Returns
        -------
        ParamUV
            The product of the parameters.
        """
        return ParamUV(self._u * other._u, self._v * other._v)

    @check_input_types
    def __truediv__(self, other: "ParamUV") -> "ParamUV":
        """
        Divides the u and v components of this ParamUV by the other ParamUV.

        Parameters
        ----------
        other : ParamUV
            The parameters to divide these parameters by.

        Returns
        -------
        ParamUV
            The quotient of the parameters.
        """
        return ParamUV(self._u / other._u, self._v / other._v)

    def __iter__(self):
        """Iterate a ``ParamUV``."""
        return iter((self.u, self.v))

    def __repr__(self) -> str:
        """Represent the ``ParamUV`` as a string."""
        return f"ParamUV(u={self.u}, v={self.v})"


class Interval:
    """
    Interval class that defines a range of values.

    Parameters
    ----------
    start : Real
        Start value of the interval.
    end : Real
        End value of the interval.
    """

    @check_input_types
    def __init__(self, start: Real, end: Real) -> None:
        """Initialize ``Interval`` class."""
        if end < start:
            raise ValueError("Start value must be less than end value")

        self._start = start
        self._end = end
        self.not_empty = True

    @property
    def start(self) -> Real:
        """Start value of the interval."""
        return self._start

    @property
    def end(self) -> Real:
        """End value of the interval."""
        return self._end

    def __eq__(self, other: object):
        """Compare two intervals."""
        if not isinstance(other, Interval):
            # don't attempt to compare against unrelated types
            return NotImplemented
        if self.is_empty() or other.is_empty():
            return self.is_empty() and other.is_empty()

        return Accuracy.equal_doubles(self.start, other.start) and Accuracy.equal_doubles(
            self.end, other.end
        )

    def is_open(self) -> bool:
        """
        If the interval is open (-inf, inf).

        Returns
        -------
        bool
            True if both ends of the interval are negative and positive infinity respectively.
        """
        return np.isneginf(self.start) and np.isinf(self.end)

    def is_closed(self) -> bool:
        """
        If the interval is closed. Neither value is inf or -inf.

        Returns
        -------
        bool
            True if neither bound of the interval is infinite.
        """
        return self.start > np.NINF and self.end < np.inf

    def is_empty(self) -> bool:
        """
        Check if the current interval is empty.

        Returns
        -------
        bool
            ``True`` when the interval is empty, ``False`` otherwise.
        """
        return not self.not_empty

    def get_span(self) -> Real:
        """
        Get the quantity contained by the interval.

        The interval must be closed.

        Returns
        -------
        Real
            The difference between the end and start of the interval.
        """
        if not self.is_closed():
            raise ValueError("Interval must be closed to get the span.")

        return self.end - self.start

    def get_relative_val(self, t: Real) -> Real:
        """
        Get an evaluation property of the interval, used in BoxUV.

        Parameters
        ----------
        t : Real
            Offset to evaluate the interval at.

        Returns
        -------
        Real
            Actual value according to the offset.
        """
        return self.start + t * self.get_span()

    def is_negative(self, tolerance: Real) -> bool:
        """
        Boolean value that indicates whether the current interval is negative.

        Parameters
        ----------
        tolerance : Real
            Accepted range because the data type of the interval could be in doubles.

        Returns
        -------
        bool
            ``True`` if the interval is negative, ``False`` otherwise.
        """
        return Accuracy.compare_with_tolerance(self.get_span(), 0, tolerance, tolerance)

    @staticmethod
    def unite(first: "Interval", second: "Interval") -> "Interval":
        """
        Get the union of two intervals.

        Parameters
        ----------
        first : Interval
            First interval.
        second : Interval
            Second interval.

        Returns
        -------
        Interval
            Union of the two intervals.
        """
        if first.is_empty():
            return second
        if second.is_empty():
            return first
        return Interval(min(first.start, second.start), max(first.end, second.end))

    def self_unite(self, other: "Interval") -> None:
        """
        Get the union of two intervals and update the current interval.

        Parameters
        ----------
        other : Interval
            Interval to unite with.
        """
        self = Interval.unite(self, other)

    @staticmethod
    def intersect(first: "Interval", second: "Interval", tolerance: Real) -> "Interval":
        """
        Get the intersection of two intervals.

        Parameters
        ----------
        first : Interval
            First interval.
        second : Interval
            Second interval.

        Returns
        -------
        Interval
            Intersection of the two intervals.
        """
        if first.is_empty() or second.is_empty():
            return None  # supposed to be empty
        intersection = Interval(max(first.start, second.start), min(first.end, second.send))
        if not intersection.is_negative(tolerance):
            return intersection
        return None  # supposed to be empty

    def self_intersect(self, other: "Interval", tolerance: Real) -> None:
        """
        Get the intersection of two intervals and update the current interval.

        Parameters
        ----------
        other : Interval
            Interval to intersect with.
        tolerance : Real
            Accepted range of error given that the interval could be in float values.
        """
        self = Interval.intersect(self, other, tolerance)

    def contains_value(self, t: Real, accuracy: Real) -> bool:
        """
        Check if the current interval contains the value ``t`` given the accuracy range.

        Parameters
        ----------
        t : Real
            Value of interest.
        accuracy : Real
            Accepted range of error given that the interval could be in float values.

        Returns
        -------
        bool
            ``True`` if the interval contains the value, ``False`` otherwise.
        """
        if self.is_empty():
            return False
        negative_abs_accuracy = -abs(accuracy)
        if self.start > self.end:
            if self.start - t < negative_abs_accuracy:
                return False
            if t - self.end < negative_abs_accuracy:
                return False
        else:
            if t - self.start < negative_abs_accuracy:
                return False
            if self.end - t < negative_abs_accuracy:
                return False
        return True

    def contains(self, t: Real) -> bool:
        """
        Check if interval contains value ``t`` using default ``length_accuracy``.

        Parameters
        ----------
        t : Real
            Value of interest.

        Returns
        -------
        bool
            ``True`` if the interval contains the value, ``False`` otherwise.
        """
        return self.contains_value(t, Accuracy.length_accuracy())

    def inflate(self, delta: Real) -> "Interval":
        """Enlarge the current interval by the given delta value."""
        if self.is_empty():
            # throw Error.InvalidMethodOnEmptyObjectException(GetType())
            raise Exception("Invalid Method On Empty Object Exception" + type(self).__name__)
        return Interval(self.start - delta, self.end + delta)

    def __repr__(self) -> str:
        """Represent the interval as a string."""
        return f"Interval(start={self.start}, end={self.end})"


@unique
class ParamForm(Enum):
    """ParamForm enum class that defines the form of a Parameterization."""

    OPEN = 1
    CLOSED = 2
    PERIODIC = 3
    OTHER = 4


@unique
class ParamType(Enum):
    """ParamType enum class that defines the type of a Parameterization."""

    LINEAR = 1
    CIRCULAR = 2
    OTHER = 3


class Parameterization:
    """
    Parameterization class describes the parameters of a specific geometry.

    Parameters
    ----------
    form : ParamForm
        Form of the parameterization.
    type : ParamType
        Type of the parameterization.
    interval : Interval
        Interval of the parameterization.
    """

    @check_input_types
    def __init__(self, form: ParamForm, type: ParamType, interval: Interval) -> None:
        """Initialize ``Parameterization`` class."""
        self._form = form
        self._type = type
        self._interval = interval

    @property
    def form(self) -> ParamForm:
        """The form of the parameterization."""
        return self._form

    @property
    def type(self) -> ParamType:
        """The type of the parameterization."""
        return self._type

    @property
    def interval(self) -> Interval:
        """The interval of the parameterization."""
        return self._interval

    def __repr__(self) -> str:
        """Represent the ``Parameterization`` as a string."""
        return f"Parameterization(form={self.form}, type={self.type}, interval={self.interval})"



# Copyright (C) 2023 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a sphere."""

from functools import cached_property

from beartype import beartype as check_input_types
from beartype.typing import Tuple, Union
import numpy as np
from pint import Quantity

from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Z
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.measurements import Distance
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
    ParamUV,
)
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.surface_evaluation import SurfaceEvaluation
from ansys.geometry.core.typing import Real, RealSequence


class Sphere(Surface):
    """
    Provides 3D sphere representation.

    Parameters
    ----------
    origin : Union[~numpy.ndarray, RealSequence, Point3D]
        Origin of the sphere.
    radius : Union[Quantity, Distance, Real]
        Radius of the sphere.
    reference : Union[~numpy.ndarray, RealSequence, UnitVector3D, Vector3D]
        X-axis direction.
    axis : Union[~numpy.ndarray, RealSequence, UnitVector3D, Vector3D]
        Z-axis direction.
    """

    @check_input_types
    def __init__(
        self,
        origin: Union[np.ndarray, RealSequence, Point3D],
        radius: Union[Quantity, Distance, Real],
        reference: Union[np.ndarray, RealSequence, UnitVector3D, Vector3D] = UNITVECTOR3D_X,
        axis: Union[np.ndarray, RealSequence, UnitVector3D, Vector3D] = UNITVECTOR3D_Z,
    ):
        """Initialize the ``Sphere`` class."""
        self._origin = Point3D(origin) if not isinstance(origin, Point3D) else origin

        self._reference = (
            UnitVector3D(reference) if not isinstance(reference, UnitVector3D) else reference
        )
        self._axis = UnitVector3D(axis) if not isinstance(axis, UnitVector3D) else axis
        if not self._reference.is_perpendicular_to(self._axis):
            raise ValueError("Circle reference (dir_x) and axis (dir_z) must be perpendicular.")

        self._radius = radius if isinstance(radius, Distance) else Distance(radius)
        if self._radius.value <= 0:
            raise ValueError("Radius must be a real positive value.")

    @property
    def origin(self) -> Point3D:
        """Origin of the sphere."""
        return self._origin

    @property
    def radius(self) -> Quantity:
        """Radius of the sphere."""
        return self._radius.value

    @property
    def dir_x(self) -> UnitVector3D:
        """X-direction of the sphere."""
        return self._reference

    @property
    def dir_y(self) -> UnitVector3D:
        """Y-direction of the sphere."""
        return self.dir_z.cross(self.dir_x)

    @property
    def dir_z(self) -> UnitVector3D:
        """Z-direction of the sphere."""
        return self._axis

    @property
    def surface_area(self) -> Quantity:
        """Surface area of the sphere."""
        return 4 * np.pi * self.radius**2

    @property
    def volume(self) -> Quantity:
        """Volume of the sphere."""
        return 4.0 / 3.0 * np.pi * self.radius**3

    @check_input_types
    def __eq__(self, other: "Sphere") -> bool:
        """Equals operator for the ``Sphere`` class."""
        return (
            self._origin == other._origin
            and self._radius == other._radius
            and self._reference == other._reference
            and self._axis == other._axis
        )

    def transformed_copy(self, matrix: Matrix44) -> "Sphere":
        """
        Create a transformed copy of the sphere based on a transformation matrix.

        Parameters
        ----------
        matrix : Matrix44
            4X4 transformation matrix to apply to the sphere.

        Returns
        -------
        Sphere
            New sphere that is the transformed copy of the original sphere.
        """
        new_point = self.origin.transform(matrix)
        new_reference = self._reference.transform(matrix)
        new_axis = self._axis.transform(matrix)
        return Sphere(
            new_point,
            self.radius,
            UnitVector3D(new_reference[0:3]),
            UnitVector3D(new_axis[0:3]),
        )

    def mirrored_copy(self) -> "Sphere":
        """
        Create a mirrored copy of the sphere along the y-axis.

        Returns
        -------
        Sphere
            New sphere that is a mirrored copy of the original sphere.
        """
        return Sphere(self.origin, self.radius, -self._reference, -self._axis)

    def evaluate(self, parameter: ParamUV) -> "SphereEvaluation":
        """
        Evaluate the sphere at the given parameters.

        Parameters
        ----------
        parameter : ParamUV
            Parameters (u,v) to evaluate the sphere at.

        Returns
        -------
        SphereEvaluation
            Resulting evaluation.
        """
        return SphereEvaluation(self, parameter)

    def project_point(self, point: Point3D) -> "SphereEvaluation":
        """
        Project a point onto the sphere and evaluate the sphere.

        Parameters
        ----------
        point : Point3D
            Point to project onto the sphere.

        Returns
        -------
        SphereEvaluation
            Resulting evaluation.
        """
        origin_to_point = point - self.origin
        x = origin_to_point.dot(self.dir_x)
        y = origin_to_point.dot(self.dir_y)
        z = origin_to_point.dot(self.dir_z)
        if np.allclose((x * x + y * y + z * z), Point3D([0, 0, 0])):
            return SphereEvaluation(self, ParamUV(0, np.pi / 2))

        u = np.arctan2(y, x)
        v = np.arctan2(z, np.sqrt(x * x + y * y))
        return SphereEvaluation(self, ParamUV(u, v))

    def parameterization(self) -> Tuple[Parameterization, Parameterization]:
        """
        Parameterization of the sphere surface as a tuple (U and V respectively).

        The U parameter specifies the longitude angle, increasing clockwise (east) about
        ``dir_z`` (right-hand corkscrew law). It has a zero parameter at ``dir_x`` and a
        period of ``2*pi``.

        The V parameter specifies the latitude, increasing north, with a zero parameter
        at the equator and a range of [-pi/2, pi/2].

        Returns
        -------
        Tuple[Parameterization, Parameterization]
            Information about how a sphere's u and v parameters are parameterized, respectively.
        """
        u = Parameterization(ParamForm.PERIODIC, ParamType.CIRCULAR, Interval(0, 2 * np.pi))
        v = Parameterization(ParamForm.CLOSED, ParamType.OTHER, Interval(-np.pi / 2, np.pi / 2))

        return (u, v)

    def contains_param(self, param_uv: ParamUV) -> bool:  # noqa: D102
        raise NotImplementedError("contains_param() is not implemented.")

    def contains_point(self, point: Point3D) -> bool:  # noqa: D102
        raise NotImplementedError("contains_point() is not implemented.")


class SphereEvaluation(SurfaceEvaluation):
    """
    Evaluate a sphere at given parameters.

    Parameters
    ----------
    sphere: ~ansys.geometry.core.shapes.surfaces.sphere.Sphere
        Sphere to evaluate.
    parameter: ParamUV
        Parameters (u, v) to evaluate the sphere at.
    """

    def __init__(self, sphere: Sphere, parameter: ParamUV) -> None:
        """Initialize the ``SphereEvaluation`` class."""
        self._sphere = sphere
        self._parameter = parameter

    @property
    def sphere(self) -> Sphere:
        """Sphere being evaluated."""
        return self._sphere

    @property
    def parameter(self) -> ParamUV:
        """Parameter that the evaluation is based upon."""
        return self._parameter

    @cached_property
    def position(self) -> Point3D:
        """
        Position of the evaluation.

        Returns
        -------
        Point3D
            Point that lies on the sphere at this evaluation.
        """
        return self.sphere.origin + self.sphere.radius.m * self.normal

    @cached_property
    def normal(self) -> UnitVector3D:
        """
        The normal to the surface.

        Returns
        -------
        UnitVector3D
            Normal unit vector to the sphere at this evaluation.
        """
        return UnitVector3D(
            np.cos(self.parameter.v) * self.__cylinder_normal
            + np.sin(self.parameter.v) * self.sphere.dir_z
        )

    @cached_property
    def __cylinder_normal(self) -> Vector3D:
        """Cylinder normal of the evaluation."""
        return (
            np.cos(self.parameter.u) * self.sphere.dir_x
            + np.sin(self.parameter.u) * self.sphere.dir_y
        )

    @cached_property
    def __cylinder_tangent(self) -> Vector3D:
        """Cylinder tangent of the evaluation."""
        return (
            -np.sin(self.parameter.u) * self.sphere.dir_x
            + np.cos(self.parameter.u) * self.sphere.dir_y
        )

    @cached_property
    def u_derivative(self) -> Vector3D:
        """
        First derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the U parameter.
        """
        return np.cos(self.parameter.v) * self.sphere.radius.m * self.__cylinder_tangent

    @cached_property
    def v_derivative(self) -> Vector3D:
        """
        First derivative with respect to the V parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the V parameter.
        """
        return self.sphere.radius.m * (
            np.cos(self.parameter.v) * self.sphere.dir_z
            - np.sin(self.parameter.v) * self.__cylinder_normal
        )

    @cached_property
    def uu_derivative(self) -> Vector3D:
        """
        Second derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            Second derivative with respect to the U parameter.
        """
        return -np.cos(self.parameter.v) * self.sphere.radius.m * self.__cylinder_normal

    @cached_property
    def uv_derivative(self) -> Vector3D:
        """
        Second derivative with respect to the U and V parameters.

        Returns
        -------
        Vector3D
            The second derivative with respect to the U and V parameters.
        """
        return -np.sin(self.parameter.v) * self.sphere.radius.m * self.__cylinder_tangent

    @cached_property
    def vv_derivative(self) -> Vector3D:
        """
        Second derivative with respect to the V parameter.

        Returns
        -------
        Vector3D
            The second derivative with respect to the V parameter.
        """
        return self.sphere.radius.m * (
            -np.sin(self.parameter.v) * self.sphere.dir_z
            - np.cos(self.parameter.v) * self.__cylinder_normal
        )

    @cached_property
    def min_curvature(self) -> Real:
        """
        Minimum curvature of the sphere.

        Returns
        -------
        Real
            Minimum curvature of the sphere.
        """
        return 1.0 / self.sphere.radius.m

    @cached_property
    def min_curvature_direction(self) -> UnitVector3D:
        """
        Minimum curvature direction.

        Returns
        -------
        UnitVector3D
            Minimum curvature direction.
        """
        return self.normal % self.max_curvature_direction

    @cached_property
    def max_curvature(self) -> Real:
        """
        Maximum curvature of the sphere.

        Returns
        -------
        Real
            Maximum curvature of the sphere.
        """
        return 1.0 / self.sphere.radius.m

    @cached_property
    def max_curvature_direction(self) -> UnitVector3D:
        """
        Maximum curvature direction.

        Returns
        -------
        UnitVector3D
            Maximum curvature direction.
        """
        return UnitVector3D(self.v_derivative)
